/*
 * crash-manager
 * Copyright (c) 2012-2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/sendfile.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdarg.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <wait.h>
#include <ctype.h>
#include <grp.h>

#include "util.h"
#include "log.h"

int write_fd(int fd, const void *buf, int len)
{
	int count;
	int total;
	total = 0;
	while (len) {
		count = write(fd, buf, len);
		if (count < 0) {
			if (errno == EINTR)
				continue;
			if (total)
				return total;
			return count;
		}
		total += count;
		buf = ((const char *)buf) + count;
		len -= count;
	}
	return total;
}

static int copy_bytes_sendfile(int destfd, int srcfd, off_t *ncopied)
{
	ssize_t nsent;
	off_t total = 0;

	do {
		nsent = sendfile(destfd, srcfd, NULL, INT32_MAX);
		if (nsent > 0)
			total += nsent;
		else if (nsent < 0 && errno == EAGAIN)
			continue;
		else
			break;
	} while (1);

	if (ncopied)
		*ncopied = total;

	return nsent == -1 ? -1 : 0;
}

static int copy_bytes_rw(int destfd, int srcfd, off_t *ncopied)
{
	ssize_t n;
	char buf[4096];
	off_t total = 0;

	do {
		n = read(srcfd, buf, sizeof buf);
		if (n < 0) {
			if (errno == EAGAIN || errno == EINTR)
				continue;
			_E("read: %m");
			return -1;
		}

		int m = write_fd(destfd, buf, n);
		if (m != n) {
			_E("failed to write data to destination fd: %m");
			return -1;
		}
		total += m;
	} while (n > 0);

	if (ncopied)
		*ncopied = total;

	return 0;
}

int copy_bytes(int destfd, int srcfd, off_t *ncopied)
{
	int r = copy_bytes_sendfile(destfd, srcfd, ncopied);
	return r == 0 ? r : copy_bytes_rw(destfd, srcfd, ncopied);
}

int copy_file(char *dst, char *src)
{
	int res;
	int sfd;
	int dfd;

	if (!src || !dst) {
		_E("Invalid argument\n");
		return -1;
	}
	sfd = open(src, O_RDONLY);
	if (sfd < 0) {
		_E("Failed to open (%s)\n", src);
		return -1;
	}
	dfd = open(dst, O_WRONLY|O_CREAT|O_EXCL, 0644);
	if (dfd < 0) {
		close(sfd);
		_SE("Failed to open (%s)\n", dst);
		return -1;
	}

	res = copy_bytes(dfd, sfd, NULL);

	close(sfd);
	close(dfd);
	return res;
}

int move_file(char *dst, char *src)
{
	if (copy_file(dst, src) < 0)
		return -1;
	if (unlink(src) < 0)
		return -1;
	return 1;
}

int dump_file_write_fd(int dfd, char *src)
{
	int res;
	int sfd;

	if (!src) {
		_E("Invalid argument\n");
		return -1;
	}
	sfd = open(src, O_RDONLY);
	if (sfd < 0) {
		_SE("Failed to open (%s)\n", src);
		return -1;
	}

	res = copy_bytes(dfd, sfd, NULL);

	close(sfd);
	return res;
}

int fsync_path(char *const path)
{
	int fd, ret;

	ret = fd = open(path, O_RDONLY);
	if (fd >= 0) {
		ret = fsync(fd);
		close(fd);
	}

	if (ret < 0)
		_E("Unable to fsync %s: %m", path);

	return ret;
}

int make_dir(const char *path, const char *name, int mode)
{
	int r = -1;

	DIR *dir = opendir(path);
	if (dir) {
		int dfd = dirfd(dir);
		r = mkdirat(dfd, name, mode);
		closedir(dir);
	}

	return r == 0 || (r == -1 && errno == EEXIST) ? 0 : -1;
}

static int remove_dir_internal(int fd)
{
	DIR *dir;
	struct dirent *de;
	int subfd, ret = 0;

	dir = fdopendir(fd);
	if (!dir)
		return -1;

	while ((de = readdir(dir))) {
		if (de->d_type == DT_DIR) {
			if (!strncmp(de->d_name, ".", 2) || !strncmp(de->d_name, "..", 3))
				continue;
			subfd = openat(fd, de->d_name, O_RDONLY | O_DIRECTORY);
			if (subfd < 0) {
				_SE("Couldn't openat %s: %d\n", de->d_name, errno);
				ret = -1;
				continue;
			}
			if (remove_dir_internal(subfd))
				ret = -1;
			close(subfd);
			if (unlinkat(fd, de->d_name, AT_REMOVEDIR) < 0) {
				_SE("Couldn't unlinkat %s: %d\n", de->d_name, errno);
				ret = -1;
			}
		} else {
			if (unlinkat(fd, de->d_name, 0) < 0) {
				_SE("Couldn't unlinkat %s: %d\n", de->d_name, errno);
				ret = -1;
			}
		}
	}
	closedir(dir);
	return ret;
}

int remove_dir(const char *path, int del_dir)
{
	int fd, ret = 0;

	if (!path)
		return -1;
	fd = open(path, O_RDONLY | O_NONBLOCK | O_DIRECTORY | O_CLOEXEC | O_NOFOLLOW);
	if (fd < 0) {
		_SE("Couldn't opendir %s: %d\n", path, errno);
		return -errno;
	}
	ret = remove_dir_internal(fd);
	close(fd);

	if (del_dir) {
		if (rmdir(path)) {
			_SE("Couldn't rmdir %s: %d\n", path, errno);
			ret = -1;
		}
	}
	return ret;
}

int get_exec_pid(const char *execpath)
{
	DIR *dp;
	struct dirent *dentry;
	int pid = -1, fd;
	int ret;
	char buf[PATH_MAX];
	char buf2[PATH_MAX];
	int len;

	dp = opendir("/proc");
	if (!dp) {
		_E("FAIL: open /proc");
		return -1;
	}

	len = strlen(execpath) + 1;

	while ((dentry = readdir(dp))) {
		if (!isdigit(dentry->d_name[0]))
			continue;

		pid = atoi(dentry->d_name);

		snprintf(buf, PATH_MAX, "/proc/%d/cmdline", pid);
		fd = open(buf, O_RDONLY);
		if (fd < 0)
			continue;
		ret = read(fd, buf2, PATH_MAX);
		close(fd);

		if (ret < 0 || ret >= PATH_MAX)
			continue;

		buf2[ret] = '\0';

		if (!strncmp(buf2, execpath, len)) {
			closedir(dp);
			return pid;
		}
	}

	errno = ESRCH;
	closedir(dp);
	return -1;
}

int get_file_count(char *path)
{
	DIR *dir;
	struct dirent *dp;
	int count = 0;

	dir = opendir(path);
	if (!dir)
		return 0;

	while ((dp = readdir(dir))) {
		const char *name = dp->d_name;
		/* always skip "." and ".." */
		if (name[0] == '.') {
			if (name[1] == 0)
				continue;
			if ((name[1] == '.') && (name[2] == 0))
				continue;
		}
		count++;
	}
	closedir(dir);
	return count;
}

off_t get_directory_usage(char *path)
{
	DIR *dir;
	struct dirent *de;
	struct stat st;
	off_t usage = 0;
	int fd = -1;

	fd = open(path, O_RDONLY|O_NONBLOCK|O_DIRECTORY|O_CLOEXEC|O_NOFOLLOW|O_NOATIME);
	if (fd < 0)
		return -1;
	dir = fdopendir(fd);
	if (!dir) {
		close(fd);
		return -1;
	}

	while ((de = readdir(dir))) {
		if (!strncmp(de->d_name, ".", 2) || !strncmp(de->d_name, "..", 3))
			continue;
		if (fstatat(fd, de->d_name, &st, AT_SYMLINK_NOFOLLOW) < 0) {
			_SE("Failed to fstatat  %s: %d\n", de->d_name, errno);
			continue;
		}
		usage += st.st_size;
	}
	closedir(dir);
	close(fd);
	return usage;
}

int log_kmsg(char *fmt, ...)
{
	int result = 1;
	va_list ap;
	FILE *file;

	file = fopen("/dev/kmsg", "w");
	if (file == NULL) {
		_E("Open /dev/kmsg error: %m");
		return 0;
	}

	va_start(ap, fmt);
	if (vfprintf(file, fmt, ap) < 0) {
		_E("Write to /dev/kmsg error: %m");
		result = 0;
	}
	va_end(ap);
	fclose(file);
	return result;
}

/**
 * @brief Check wchan of thread
 *
 * @param pid PID of the inspected process
 * @param tid TID of the thread to check
 */
static int check_thread_wchan(int pid, int tid)
{
	int fd, cnt;
	char path[PATH_MAX], buf[100];

	snprintf(path, sizeof(path), "/proc/%d/task/%d/wchan", pid, tid);
	fd = open(path, O_RDONLY);
	if (fd == -1) {
		_E("cannot open %s: %m\n", path);
		return -errno;
	}
	cnt = read(fd, buf, sizeof(buf));
	if (cnt == -1 || cnt == sizeof(buf)) {
		_E("read %s error: %m\n", path);
		close(fd);
		return -errno;
	}
	buf[cnt] = 0;
	close(fd);

	if (strncmp("do_coredump", buf, sizeof(buf)) == 0 || strncmp("pipe_wait", buf, sizeof(buf)) == 0)
		return tid;
	else
		return 0;
}

/**
 * @brief Find crashed tid if tid was not offered
 *
 * @param pid PID of the inspected process
 */
int find_crash_tid(int pid)
{
	int threadnum = 1;
	int crash_tid = -1;
	DIR *dir;
	struct dirent *entry;
	char task_path[PATH_MAX];
	struct stat sb;

	snprintf(task_path, sizeof(task_path), "/proc/%d/task", pid);
	if (stat(task_path, &sb) == -1) {
		_E("no such file: %s", task_path);
		return -1;
	}

	threadnum = sb.st_nlink - 2;

	if (threadnum > 1) {
		dir = opendir(task_path);
		if (!dir) {
			_E("cannot open %s\n", task_path);
			return -1;
		} else {
			while ((entry = readdir(dir)) != NULL) {
				if (strcmp(entry->d_name, ".") == 0 ||
				    strcmp(entry->d_name, "..") == 0)
					continue;
				crash_tid = check_thread_wchan(pid,
							       atoi(entry->d_name));
				if (crash_tid > 0)
					break;
			}
			closedir(dir);
			return crash_tid;
		}
	} else if (threadnum == 1) {
		return pid;
	}
	return -1;
}

bool filter_drop_trailing_whitespace(char *buf, int size, int datalen)
{
	assert(buf);
	assert(datalen <= size);

	bool filtered = false;

	for (int i = datalen; i >= 0 && isspace(buf[i]); --i) {
		buf[i] = '\0';
		filtered = true;
	}

	return filtered;
}

bool read_proc_file(pid_t pid, const char * const file, char *outbuf, size_t bufsize, charp0filter filter)
{
	assert(file);
	assert(outbuf);
	assert(bufsize > 0);

	char path[PATH_MAX];

	snprintf(path, sizeof(path), "/proc/%d/%s", pid, file);

	int fd = open(path, O_RDONLY);
	if (fd < 0) {
		_E("Failed to open %s: %m\n", path);
		return false;
	}

	ssize_t ret = read(fd, outbuf, bufsize - 1);
	if (ret < 0) {
		_E("Failed to read %s: %m\n", path);
		goto err_close;
	}
	outbuf[ret] = '\0';

	close(fd);

	if (ret > 0 && filter)
		(void)filter(outbuf, bufsize, ret - 1);

	return true;

err_close:
	close(fd);
	return false;
}

bool get_exe_path(pid_t pid, char *outbuf, size_t outbuf_len)
{
	assert(outbuf);
	assert(outbuf_len > 0);

	char exe_link[PATH_MAX];

	snprintf(exe_link, sizeof(exe_link), "/proc/%d/exe", pid);

	ssize_t ret = readlink(exe_link, outbuf, outbuf_len - 1);
	if (ret <= 0) {
		_E("Failed to read link %s: %m", exe_link);
		return false;
	}
	outbuf[ret] = '\0';

	if (access(outbuf, F_OK) == -1) {
		_E("Unable to access %s: %m", outbuf);
		return false;
	}
	return true;
}

/* This function is supposed to accept same data as passed to execve
 * (argv and envp), which can be arrays of strings as well as NULL
 * pointer.
 */
char* concatenate(char *const vec[])
{
	size_t length = 0;
	for (char *const *p = vec; p && *p; p++)
		length += strlen(*p) + 1;

	if (length == 0)
		return strdup("");

	char *str = (char *)malloc(length);
	if (!str)
		return NULL;

	char *destp = str;
	char *const *vecp = vec;
	while (*vecp) {
		destp = stpcpy(destp, *(vecp++));
		if (*vecp)
			destp = stpcpy(destp, " ");
	}

	return str;
}

bool string_ends_with(const char *string, const char *suffix)
{
	const size_t string_len = strlen(string);
	const size_t suffix_len = strlen(suffix);

	return (string_len >= suffix_len) && !strcmp(string + string_len - suffix_len, suffix);
}

bool file_exists(const char *path)
{
	struct stat buf;
	return stat(path, &buf) == 0;
}

/**
 * @}
 */
